# -*- coding: utf-8 -*-
"""
Configuration settings. Can read additional/overridden options from INI file,
supporting any JSON-serializable datatype.

INI file can contain a [DEFAULT] section for default settings, and additional
sections overriding the default for different environments. Example:
-----------------
[DEFAULT]
# single-line comments can start with # or ;
ServerIP = my.server.domain
ServerPort = 80
SampleJSON = {"a": false, "b": [0.1, 0.2]}

[DEV]
ServerIP = 0.0.0.0

save() retains only the DEFAULT section, and writes only values diverging from
the declared ones in source code. File is deleted if all values are at default.

@author      Erki Suurjaak
@created     26.03.2015
@modified    25.03.2024
------------------------------------------------------------------------------
"""
import ast
try: import ConfigParser as configparser # Py2
except ImportError: import configparser  # Py3
import datetime
import io
import json
import logging
import os
import re
import sys

"""Program title, version number and version date."""
Title = "InputScope"
Version = "1.10"
VersionDate = "25.03.2024"

"""TCP port of the web user interface."""
WebHost = "localhost"
WebPort = 8099
WebUrl = "http://%s:%s" % (WebHost, WebPort)

HomepageUrl = "https://github.com/suurjaak/InputScope"

"""Size of the heatmaps, in pixels."""
MouseHeatmapSize    = (640, 360)
KeyboardHeatmapSize = (680, 180)

"""Default desktop size for scaling, if not available from system, in pixels."""
DefaultScreenSize = (1920, 1080)

"""Whether mouse or keyboard logging is enabled."""
MouseEnabled    = True
KeyboardEnabled = True

"""Individual mouse event logging flags."""
MouseMovesEnabled   = True
MouseClicksEnabled  = True
MouseScrollsEnabled = True

"""Individual key event logging flags."""
KeyboardKeysEnabled   = True
KeyboardCombosEnabled = True

"""Mapping inputs to event types."""
InputEvents = {
    "mouse":    ("moves", "clicks", "scrolls"),
    "keyboard": ("keys", "combos")
}
"""Mapping inputs and event types to config flag names."""
InputFlags = {
    "mouse"   : "MouseEnabled",
    "keyboard": "KeyboardEnabled",

    "moves"   : "MouseMovesEnabled",
    "clicks"  : "MouseClicksEnabled",
    "scrolls" : "MouseScrollsEnabled",
    "keys"    : "KeyboardKeysEnabled",
    "combos"  : "KeyboardCombosEnabled",
}

"""Extra configured keys, as {virtual keycode: "key name"}."""
CustomKeys = {}

"""Maximum keypress interval to count as one typing session, in seconds."""
KeyboardSessionMaxDelta = 3

"""Whether to ignore repeated keyboard events from long keypresses."""
KeyboardStickyEnabled = True

"""Maximum interval between linear move events for event reduction, in seconds (0 disables)."""
MouseMoveJoinInterval = 0.5

"""Fuzz radius for linear move events for event reduction, in heatmap pixels."""
MouseMoveJoinRadius = 5

"""Maximum interval between scroll events for event reduction, in seconds (0 disables)."""
MouseScrollJoinInterval = 0.5

"""Interval between writing events to database, in seconds."""
EventsWriteInterval = 5

"""Interval between checking and saving screen size, in seconds."""
ScreenSizeInterval = 10

"""Maximum number of events to use for statistics page."""
MaxEventsForStats = 1000 * 1000

"""Maximum number of events to replay on statistics page."""
MaxEventsForReplay = 100 * 1000

"""Maximum number of events to queue for database insertion, excess is discarded."""
MaxEventsForQueue = 1000

"""Maximum number of sessions listed in tray menu."""
MaxSessionsInMenu = 20

"""Maxinum number of most common keys/combos for applications on statistics page."""
MaxTopKeysForPrograms = 5

"""
List of screen areas to monitor for mouse events if not all,
as [[x, y, w, h], ] or [[screen index, [x, y, w, h]], ]; coordinates
can be given as pixels, or as percentages of screen size (decimal fractions 0..1).
"""
MouseRegionsOfInterest = []

"""
List of screen areas to ignore for mouse events,
as [[x, y, w, h], ] or [[screen index, [x, y, w, h]], ]; coordinates
can be given as pixels, or as percentages of screen size (decimal fractions 0..1).
"""
MouseRegionsOfDisinterest = []

"""Physical length of a pixel, in meters."""
PixelLength = 0.00024825

"""
Applications to ignore for inputs events,
as {executable path: [] if all inputs else [input or event type, ]}.

Path can be absolute or relative like "C:\Python\python.exe" or "python.exe",
and can contain wildcards like "python*".
"""
ProgramBlacklist = {}

"""
Applications to monitor inputs from if not all,
as {executable path: [] if all inputs else [input or event type, ]}.

Path can be absolute or relative like "C:\Python\python.exe" or "python.exe",
and can contain wildcards like "python*".
"""
ProgramWhitelist = {}

"""Whether active application logging, filtering and statistics are enabled."""
ProgramsEnabled = True

"""Ordered mapping of tables to input types."""
InputTables = [("mouse", ["moves", "clicks", "scrolls"]), ("keyboard", ["keys", "combos"])]

"""Key positions in keyboard heatmap."""
KeyPositions = {
  "Escape": (12, 12),
  "F1": (72, 12),
  "F2": (102, 12),
  "F3": (132, 12),
  "F4": (162, 12),
  "F5": (206, 12),
  "F6": (236, 12),
  "F7": (266, 12),
  "F8": (296, 12),
  "F9": (338, 12),
  "F10": (368, 12),
  "F11": (398, 12),
  "F12": (428, 12),
  "PrintScreen": (472, 12),
  "ScrollLock": (502, 12),
  "Pause": (532, 12),
  "Break": (532, 12),

  "Oem_7": (12, 56),
  "1": (44, 56),
  "2": (74, 56),
  "3": (104, 56),
  "4": (134, 56),
  "5": (164, 56),
  "6": (192, 56),
  "7": (222, 56),
  "8": (252, 56),
  "9": (281, 56),
  "0": (311, 56),
  "Oem_Minus": (340, 56),
  "Oem_Plus": (371, 56),
  "Backspace": (414, 56),

  "Tab": (24, 84),
  "Q": (60, 84),
  "W": (90, 84),
  "E": (120, 84),
  "R": (150, 84),
  "T": (180, 84),
  "Y": (210, 84),
  "U": (240, 84),
  "I": (270, 84),
  "O": (300, 84),
  "P": (330, 84),
  "Oem_3": (360, 84),
  "Oem_4": (390, 84),
  "Enter": (426, 96),

  "CapsLock": (25, 111),
  "A": (68, 111),
  "S": (98, 111),
  "D": (128, 111),
  "F": (158, 111),
  "G": (188, 111),
  "H": (218, 111),
  "J": (248, 111),
  "K": (278, 111),
  "L": (308, 111),
  "Oem_1": (338, 111),
  "Oem_2": (368, 111),
  "Oem_5": (394, 111),

  "Lshift": (19, 138),
  "Oem_102": (50, 138),
  "Z": (80, 138),
  "X": (110, 138),
  "C": (140, 138),
  "V": (170, 138),
  "B": (200, 138),
  "N": (230, 138),
  "M": (260, 138),
  "Oem_Comma": (290, 138),
  "Oem_Period": (320, 138),
  "Oem_6": (350, 138),
  "Rshift": (404, 138),

  "Lcontrol": (19, 166),
  "Lwin": (54, 166),
  "Alt": (89, 166),
  "Space": (201, 166),
  "AltGr": (315, 166),
  "Rwin": (350, 166),
  "Menu": (384, 166),
  "Rcontrol": (424, 166),

  "Up": (504, 138),
  "Left": (474, 166),
  "Down": (504, 166),
  "Right": (534, 166),

  "Insert": (474, 56),
  "Home": (504, 56),
  "PageUp": (534, 56),
  "Delete": (474, 84),
  "End": (504, 84),
  "PageDown": (534, 84),

  "NumLock": (576, 56),
  "Numpad-Divide": (605, 56),
  "Numpad-Multiply": (634, 56),
  "Numpad-Subtract": (664, 56),
  "Numpad-Add": (664, 98),
  "Numpad-Enter": (664, 152),
  "Numpad0": (590, 166),
  "Numpad1": (576, 138),
  "Numpad2": (605, 138),
  "Numpad3": (634, 138),
  "Numpad4": (576, 111),
  "Numpad5": (605, 111),
  "Numpad6": (634, 111),
  "Numpad7": (576, 84),
  "Numpad8": (605, 84),
  "Numpad9": (634, 84),
  "Numpad-Insert": (590, 166),
  "Numpad-Decimal": (634, 166),
  "Numpad-Delete": (634, 166),
  "Numpad-End": (576, 138),
  "Numpad-Down": (605, 138),
  "Numpad-PageDown": (634, 138),
  "Numpad-Left": (576, 111),
  "Numpad-Clear": (605, 111),
  "Numpad-Right": (634, 111),
  "Numpad-Home": (576, 84),
  "Numpad-Up": (605, 84),
  "Numpad-PageUp": (634, 84),
}

"""Whether web modules and templates are automatically reloaded on change."""
WebAutoReload = False

"""Whether web server is quiet or echoes access log."""
WebQuiet = False

"""Whether running as a pyinstaller executable."""
Frozen = getattr(sys, "frozen", False)
if Frozen:
    ExecutablePath = ShortcutIconPath = os.path.abspath(sys.executable)
    ApplicationPath = os.path.dirname(ExecutablePath)
    RootPath = os.path.join(os.environ.get("_MEIPASS2", getattr(sys, "_MEIPASS", "")))
    DbPath = os.path.join(ApplicationPath, "%s.db" % Title.lower())
    ConfigPath = os.path.join(ApplicationPath, "%s.ini" % Title.lower())
else:
    RootPath = ApplicationPath = os.path.dirname(os.path.abspath(__file__))
    ExecutablePath = os.path.join(RootPath, "main.py")
    ShortcutIconPath = os.path.join(RootPath, "static", "icon.ico")
    DbPath = os.path.join(RootPath, "var", "%s.db" % Title.lower())
    ConfigPath = os.path.join(ApplicationPath, "var", "%s.ini" % Title.lower())

"""Path for static web content, like images and JavaScript files."""
StaticPath = os.path.join(RootPath, "static")

"""Path for HTML templates."""
TemplatePath = os.path.join(RootPath, "views")

"""Path for application icon file."""
IconPath = os.path.join(StaticPath, "icon.ico")

"""Path for licences of bundled open-source software."""
LicensePath = os.path.join(StaticPath, "3rd-party licenses.txt") if Frozen else None

"""SQL template for trigger to update day counts."""
TriggerTemplate = """
CREATE TRIGGER IF NOT EXISTS on_insert_{0} AFTER INSERT ON {0}
BEGIN
  INSERT OR IGNORE INTO counts (type, day, count) VALUES ('{0}', NEW.day, 0);
  UPDATE counts SET count = count + 1 WHERE type = '{0}' AND day = NEW.day;
END;"""

"""SQL template for day field index."""
DayIndexTemplate = "CREATE INDEX IF NOT EXISTS idx_{0}_day ON {0} (day)"

"""Statements to execute in database at startup, like CREATE TABLE."""
DbStatements = (
    "CREATE TABLE IF NOT EXISTS moves (id INTEGER NOT NULL PRIMARY KEY, day DATE, stamp REAL, x INTEGER, y INTEGER, display INTEGER DEFAULT 0)",
    "CREATE TABLE IF NOT EXISTS clicks (id INTEGER NOT NULL PRIMARY KEY, day DATE, stamp REAL, x INTEGER, y INTEGER, button INTEGER, display INTEGER DEFAULT 0)",
    "CREATE TABLE IF NOT EXISTS scrolls (id INTEGER NOT NULL PRIMARY KEY, day DATE, stamp REAL, x INTEGER, y INTEGER, dx INTEGER DEFAULT 0, dy INTEGER DEFAULT 0, display INTEGER DEFAULT 0)",
    "CREATE TABLE IF NOT EXISTS keys (id INTEGER NOT NULL PRIMARY KEY, day DATE, stamp REAL, key TEXT, realkey TEXT)",
    "CREATE TABLE IF NOT EXISTS combos (id INTEGER NOT NULL PRIMARY KEY, day DATE, stamp REAL, key TEXT, realkey TEXT)",
    "CREATE TABLE IF NOT EXISTS app_events (id INTEGER NOT NULL PRIMARY KEY, dt TIMESTAMP DEFAULT (DATETIME('now', 'localtime')), type TEXT)",
    "CREATE TABLE IF NOT EXISTS screen_sizes (id INTEGER NOT NULL PRIMARY KEY, dt TIMESTAMP DEFAULT (DATETIME('now', 'localtime')), x INTEGER, y INTEGER, w INTEGER, h INTEGER, display INTEGER)",
    "CREATE TABLE IF NOT EXISTS counts (id INTEGER NOT NULL PRIMARY KEY, type TEXT, day DATETIME, count INTEGER, UNIQUE(type, day))",
    "CREATE TABLE IF NOT EXISTS sessions (id INTEGER NOT NULL PRIMARY KEY, name TEXT, day1 DATETIME, day2 DATETIME, start REAL, end REAL)",
    "CREATE TABLE IF NOT EXISTS programs (id INTEGER PRIMARY KEY AUTOINCREMENT NOT NULL, path TEXT NOT NULL)",
) + tuple(TriggerTemplate .format(t) for _, tt in InputTables for t in tt
) + tuple(DayIndexTemplate.format(t) for _, tt in InputTables for t in tt)

"""
Statements to update database to new schema, as {(table, column to check if exists): [ALTER SQLs]}.
"""
DbUpdateStatements = [
    # v1.3+
    [("moves",   "display"), ["ALTER TABLE moves ADD COLUMN display INTEGER DEFAULT 0"]],
    [("clicks",  "display"), ["ALTER TABLE clicks ADD COLUMN display INTEGER DEFAULT 0"]],
    [("scrolls", "display"), ["ALTER TABLE scrolls ADD COLUMN display INTEGER DEFAULT 0"]],
    [("screen_sizes", "display"),  [
        "ALTER TABLE screen_sizes RENAME COLUMN x TO w",
        "ALTER TABLE screen_sizes RENAME COLUMN y TO h",
        "ALTER TABLE screen_sizes ADD COLUMN x INTEGER DEFAULT 0",
        "ALTER TABLE screen_sizes ADD COLUMN y INTEGER DEFAULT 0",
        "ALTER TABLE screen_sizes ADD COLUMN display INTEGER DEFAULT 0"]],
    [("scrolls", "dy"),  [
        "ALTER TABLE scrolls RENAME COLUMN wheel TO dy"]],
    [("scrolls", "dx"),  [
        "ALTER TABLE scrolls ADD COLUMN dx INTEGER DEFAULT 0"]],
    # v1.8+ 
    [("clicks",  "fk_program"), ["ALTER TABLE clicks ADD COLUMN fk_program INTEGER"]],
    [("combos",  "fk_program"), ["ALTER TABLE combos ADD COLUMN fk_program INTEGER"]],
    [("keys",    "fk_program"), ["ALTER TABLE keys ADD COLUMN fk_program INTEGER"]],
    [("moves",   "fk_program"), ["ALTER TABLE moves ADD COLUMN fk_program INTEGER"]],
    [("scrolls", "fk_program"), ["ALTER TABLE scrolls ADD COLUMN fk_program INTEGER"]],
]

"""List of attribute names that are always saved to ConfigFile."""
FileDirectives = ["CustomKeys", "DefaultScreenSize", "EventsWriteInterval", "MaxEventsForStats",
    "MaxEventsForReplay", "KeyboardEnabled", "KeyboardKeysEnabled", "KeyboardCombosEnabled",
    "KeyboardStickyEnabled", "MouseEnabled", "MouseMovesEnabled", "MouseClicksEnabled",
    "MouseScrollsEnabled", "MouseHeatmapSize", "MouseMoveJoinInterval", "MouseMoveJoinRadius",
    "MouseScrollJoinInterval", "MouseRegionsOfInterest", "MouseRegionsOfDisinterest", "PixelLength",
    "ProgramBlacklist", "ProgramWhitelist", "ProgramsEnabled", "ScreenSizeInterval", "WebPort",
]

try: text_types = (str, unicode)       # Py2
except Exception: text_types = (str, ) # Py3

def init(filename=ConfigPath, create=True):
    """Loads INI configuration into this module's attributes; creates file if not present."""
    section, parts = "DEFAULT", filename.rsplit(":", 1)
    if len(parts) > 1 and os.path.isfile(parts[0]): filename, section = parts
    if not os.path.isfile(filename):
        if create: save()
        return

    vardict, parser = globals(), configparser.RawConfigParser()
    parser.optionxform = str # Force case-sensitivity on names
    try:
        def parse_value(raw):
            try: return json.loads(raw) # Try to interpret as JSON
            except ValueError:
                try: return ast.literal_eval(raw) # JSON failed, fall back to eval
                except (SyntaxError, ValueError): raw # Fall back to raw
        with open(filename, "r") as f:
            txt = f.read()
        try: txt = txt.decode()
        except Exception: pass
        if not re.search("\\[\\w+\\]", txt): txt = "[DEFAULT]\n" + txt
        reader = getattr(parser, "read_file", getattr(parser, "readfp", None))  # Py3/Py2
        reader(io.StringIO(txt), filename)
        for k, v in parser.items(section): vardict[k] = parse_value(v)
    except Exception:
        logging.warn("Error reading config from %s.", filename, exc_info=True)
    validate()


def save(filename=ConfigPath):
    """Saves this module's changed attributes to INI configuration."""
    default_values = defaults()
    parser = configparser.RawConfigParser()
    parser.optionxform = str # Force case-sensitivity on names
    try:
        save_types = text_types + (int, float, tuple, list, dict, type(None))
        for k, v in sorted(globals().items()):
            if not isinstance(v, save_types) or k.startswith("_") or k not in default_values \
            or k not in FileDirectives and default_values[k] == v: continue # for k, v
            try: parser.set("DEFAULT", k, json.dumps(v))
            except Exception: pass
        with open(filename, "w") as f:
            f.write("# %s %s configuration written on %s.\n" % (Title, Version,
                    datetime.datetime.now().strftime("%Y-%m-%d %H:%M:%S")))
            parser.write(f)
    except Exception:
        logging.warn("Error writing config to %s.", filename, exc_info=True)


def validate():
    """Validates configuration values, discarding invalids."""
    global CustomKeys
    default_values = defaults()
    for k, v in sorted(globals().items()):
        v0 = default_values.get(k)
        if isinstance(v, list) and isinstance(v0, tuple):
            globals()[k] = tuple(v)
    try:
        keys, _ = CustomKeys.copy(), CustomKeys.clear()
        for k, v in keys.items():
            try: CustomKeys[int(k)] = v
            except Exception:
                try: CustomKeys[int(k, 16)] = v
                except Exception: pass
    except Exception: CustomKeys = defaults()["CustomKeys"]


def defaults(values={}):
    """Returns a once-assembled dict of this module's storable attributes."""
    if values: return values
    save_types = text_types + (int, float, tuple, list, dict, type(None))
    for k, v in globals().items():
        if isinstance(v, save_types) and not k.startswith("_"): values[k] = v
    return values


defaults() # Store initial values to compare on saving
